from phylogenie.treesimulator.events.base import Event, TimedEvent
from phylogenie.treesimulator.events.unbounded_population import (
    EXPOSED_STATE,
    INFECTIOUS_STATE,
    SUPERSPREADER_STATE,
    Birth,
    Death,
    Migration,
    Sampling,
    TimedSampling,
    UnboundedPopulationEvent,
    UnboundedPopulationTimedEvent,
    get_BD_events,
    get_BDEI_events,
    get_BDSS_events,
    get_canonical_events,
    get_epidemiological_events,
    get_FBD_events,
)

__all__ = [
    "EXPOSED_STATE",
    "INFECTIOUS_STATE",
    "SUPERSPREADER_STATE",
    "Event",
    "TimedEvent",
    "Birth",
    "Death",
    "Migration",
    "Sampling",
    "TimedSampling",
    "UnboundedPopulationEvent",
    "UnboundedPopulationTimedEvent",
    "get_BD_events",
    "get_BDEI_events",
    "get_BDSS_events",
    "get_canonical_events",
    "get_epidemiological_events",
    "get_FBD_events",
]



from phylogenie.treesimulator.models.base import Model
from phylogenie.treesimulator.models.unbounded_population import (
    UnboundedPopulationModel,
)

__all__ = ["Model", "UnboundedPopulationModel"]



import time
from collections import defaultdict
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Callable, TypeVar

import joblib
import numpy as np
import pandas as pd
from numpy.random import default_rng
from tqdm import tqdm

from phylogenie.io import dump_newick
from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator.events import Event, TimedEvent
from phylogenie.treesimulator.models import Model

M = TypeVar("M", bound="Model")


def simulate_tree(
    events: Sequence[Event[M]],
    model: M,
    timed_events: Sequence[TimedEvent[M]] | None = None,
    n_leaves: int | None = None,
    max_time: float | None = None,
    seed: int | None = None,
    timeout: float | None = None,
    acceptance_criterion: Callable[[TreeNode], bool] | None = None,
    logs: dict[str, Callable[[TreeNode], Any]] | None = None,
) -> tuple[TreeNode, dict[str, Any]]:
    if (max_time is None) == (n_leaves is None):
        raise ValueError("Exactly one of max_time or n_leaves must be specified.")

    events_at_time: dict[float, list[TimedEvent[M]]] = defaultdict(list)
    if timed_events is not None:
        for te in timed_events:
            for t in te.times:
                events_at_time[t].append(te)

    rng = default_rng(seed)
    start_clock = time.perf_counter()
    while True:
        model.init()
        current_time = 0.0
        change_times = sorted(set(t for e in events for t in e.rate.change_times))
        next_change_time = change_times.pop(0) if change_times else None
        timed_event_times = sorted(events_at_time.keys())
        next_timed_event_time = timed_event_times.pop(0) if timed_event_times else None

        while n_leaves is None or model.tree_size < n_leaves:
            if timeout is not None and time.perf_counter() - start_clock > timeout:
                raise TimeoutError("Simulation timed out.")

            propensities = [e.get_propensity(model, current_time) for e in events]
            total_propensity = sum(propensities)
            next_event_time = (
                current_time + rng.exponential(1 / total_propensity)
                if total_propensity
                else None
            )

            if (
                next_change_time is not None
                and (next_event_time is None or next_change_time < next_event_time)
                and (
                    next_timed_event_time is None
                    or next_change_time <= next_timed_event_time
                )
                and (max_time is None or next_change_time < max_time)
            ):  # The next event is a rate change
                current_time = next_change_time
                next_change_time = change_times.pop(0) if change_times else None
            elif (
                next_timed_event_time is not None
                and (next_event_time is None or next_timed_event_time < next_event_time)
                and (max_time is None or next_timed_event_time <= max_time)
            ):  # The next event is a timed event
                current_time = next_timed_event_time
                for te in events_at_time[current_time]:
                    te.apply(model, current_time, rng)
                next_timed_event_time = (
                    timed_event_times.pop(0) if timed_event_times else None
                )
            elif next_event_time is not None and (
                max_time is None or next_event_time < max_time
            ):  # The next event is a stochastic event
                current_time = next_event_time
                event_idx = np.searchsorted(
                    np.cumsum(propensities) / total_propensity, rng.random()
                )
                event = events[int(event_idx)]
                event.apply(model, current_time, rng)
            else:  # No more events can occur
                break

        # If the simulation stopped because no more events could occur, restart
        if n_leaves is not None and model.tree_size < n_leaves:
            continue

        tree = model.get_tree()

        if (
            tree is None
            or acceptance_criterion is not None
            and not acceptance_criterion(tree)
        ):
            continue

        metadata: dict[str, Any] = {}
        if logs is not None:
            for key, func in logs.items():
                metadata[key] = func(tree)

        return (tree, metadata)


def generate_trees(
    output_dir: str | Path,
    n_trees: int,
    events: Sequence[Event[M]],
    model: M,
    timed_events: Sequence[TimedEvent[M]] | None = None,
    n_leaves: int | None = None,
    max_time: float | None = None,
    node_features: Mapping[str, str] | None = None,
    seed: int | None = None,
    n_jobs: int = -1,
    timeout: float | None = None,
    acceptance_criterion: Callable[[TreeNode], bool] | None = None,
    logs: dict[str, Callable[[TreeNode], Any]] | None = None,
) -> pd.DataFrame:
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    if output_dir.exists():
        raise FileExistsError(f"Output directory {output_dir} already exists")
    output_dir.mkdir(parents=True)

    def _simulate_tree(i: int, seed: int) -> dict[str, Any]:
        while True:
            try:
                tree, metadata = simulate_tree(
                    events=events,
                    n_leaves=n_leaves,
                    max_time=max_time,
                    model=model,
                    timed_events=timed_events,
                    seed=seed,
                    timeout=timeout,
                    acceptance_criterion=acceptance_criterion,
                    logs=logs,
                )
                metadata["file_id"] = i
                if node_features is not None:
                    for name, feature in node_features.items():
                        mapping = getattr(tree, feature)
                        for node in tree:
                            node[name] = mapping[node]
                dump_newick(tree, output_dir / f"{i}.nwk")
                return metadata
            except TimeoutError as e:
                print(f"{e}. Retrying with a different seed...")
            seed += 1

    rng = default_rng(seed)
    jobs = joblib.Parallel(n_jobs=n_jobs, return_as="generator_unordered")(
        joblib.delayed(_simulate_tree)(i=i, seed=int(rng.integers(2**32)))
        for i in range(n_trees)
    )

    return pd.DataFrame(
        [md for md in tqdm(jobs, f"Generating trees in {output_dir}...", n_trees)]
    )



from abc import ABC, abstractmethod
from collections.abc import Sequence
from math import comb
from typing import Generic, TypeVar

from numpy.random import Generator

from phylogenie.skyline import SkylineParameterLike, skyline_parameter
from phylogenie.treesimulator.models import Model

M = TypeVar("M", bound="Model")


class Event(ABC, Generic[M]):
    def __init__(self, reactants: dict[str, int], rate: SkylineParameterLike):
        self.reactants = reactants
        self.rate = skyline_parameter(rate)
        if any(v < 0 for v in self.rate.value):
            raise ValueError("Event rates must be non-negative.")

    def draw_reactants(self, model: M, rng: Generator) -> dict[str, list[int]]:
        return {
            state: rng.choice(
                model.get_individuals(state), size=count, replace=False
            ).tolist()
            for state, count in self.reactants.items()
        }

    def get_propensity(self, model: M, time: float) -> float:
        propensity = self.rate.get_value_at_time(time)
        for state, count in self.reactants.items():
            propensity *= comb(model.count_individuals(state), count)
        return propensity

    @abstractmethod
    def apply(self, model: M, time: float, rng: Generator): ...


class SingleReactantEvent(Event[M]):
    def __init__(self, state: str, rate: SkylineParameterLike):
        super().__init__({state: 1}, rate)
        self.state = state

    def draw_individual(self, model: M, rng: Generator) -> int:
        return rng.choice(model.get_individuals(self.state))


class TimedEvent(ABC, Generic[M]):
    def __init__(self, times: Sequence[float]):
        if any(t < 0 for t in times):
            raise ValueError("Event times must be non-negative.")
        self.times = sorted(times)

    @abstractmethod
    def apply(self, model: M, time: float, rng: Generator): ...



from abc import ABC, abstractmethod
from collections import defaultdict

from phylogenie._mixins import MetadataMixin
from phylogenie.tree_node import TreeNode


class Model(MetadataMixin, ABC):
    def init(self):
        self.clear()
        self._next_individual_id = 0
        self._population: dict[str, set[int]] = defaultdict(set)
        self._states: dict[int, str] = {}

    def create_new_individual(self, state: str) -> int:
        self._next_individual_id += 1
        self._population[state].add(self._next_individual_id)
        self._states[self._next_individual_id] = state
        return self._next_individual_id

    def remove_individual(self, individual: int) -> None:
        state = self._states[individual]
        self._population[state].remove(individual)
        del self._states[individual]

    def migrate_individual(self, individual: int, new_state: str) -> None:
        old_state = self._states[individual]
        self._population[old_state].remove(individual)
        self._states[individual] = new_state
        self._population[new_state].add(individual)

    def get_individuals(self, state: str | None = None) -> list[int]:
        if state is None:
            return list(self._states)
        return list(self._population[state])

    def count_individuals(self, state: str | None = None) -> int:
        return len(self.get_individuals(state))

    @property
    @abstractmethod
    def tree_size(self) -> int: ...

    @abstractmethod
    def get_tree(self) -> TreeNode | None: ...



from phylogenie.tree_node import TreeNode
from phylogenie.treesimulator.models.base import Model


def _get_node_name(node_id: int, state: str) -> str:
    return f"{node_id}|{state}"


def get_node_state(node_name: str) -> str:
    if "|" not in node_name:
        raise ValueError(
            f"Invalid node name: {node_name} (expected format 'id|state')."
        )
    return node_name.split("|")[-1]


class UnboundedPopulationModel(Model):
    def __init__(self, init_state: str):
        super().__init__()
        self._init_state = init_state
        self.init()

    def init(self):
        super().init()
        self._next_node_id = 0
        self._sampled: set[str] = set()
        self._times: dict[TreeNode, float] = {}
        self._nodes: dict[int, TreeNode] = {}
        seed_individual = self._get_new_individual(self._init_state)
        self._tree = self._nodes[seed_individual]

    @property
    def tree_size(self) -> int:
        return len(self._sampled)

    def _get_new_node(self, state: str) -> TreeNode:
        self._next_node_id += 1
        node = TreeNode(_get_node_name(self._next_node_id, state))
        return node

    def _get_new_individual(self, state: str) -> int:
        new_id = self.create_new_individual(state)
        node = self._get_new_node(state)
        self._nodes[new_id] = node
        return new_id

    def _set_branch_length(self, node: TreeNode, time: float) -> None:
        if node.branch_length is not None:
            raise ValueError(f"Branch length of node {node.name} is already set.")
        parent_time = 0.0 if node.parent is None else self._times[node.parent]
        node.branch_length = time - parent_time
        self._times[node] = time

    def _stem(self, individual: int, time: float) -> None:
        node = self._nodes[individual]
        self._set_branch_length(node, time)
        stem_node = self._get_new_node(self._states[individual])
        node.add_child(stem_node)
        self._nodes[individual] = stem_node

    def remove(self, individual: int, time: float) -> None:
        self.remove_individual(individual)
        self._set_branch_length(self._nodes[individual], time)

    def migrate(self, individual: int, new_state: str, time: float) -> None:
        self.migrate_individual(individual, new_state)
        self._stem(individual, time)

    def birth_from(self, parent_individual: int, new_state: str, time: float) -> int:
        new_individual = self._get_new_individual(new_state)
        parent_node = self._nodes[parent_individual]
        new_node = self._nodes[new_individual]
        parent_node.add_child(new_node)
        self._stem(parent_individual, time)
        return new_individual

    def sample(self, individual: int, time: float, removal: bool) -> None:
        if removal:
            self._sampled.add(self._nodes[individual].name)
            self.remove(individual, time)
        else:
            sample_node = self._get_new_node(self._states[individual])
            sample_node.branch_length = 0.0
            self._sampled.add(sample_node.name)
            self._nodes[individual].add_child(sample_node)
            self._stem(individual, time)

    def get_tree(self) -> TreeNode | None:
        tree = self._tree.copy()
        for node in list(tree.iter_postorder()):
            if node.name not in self._sampled and not node.children:
                if node.parent is None:
                    return None
                else:
                    node.parent.remove_child(node)
            elif len(node.children) == 1:
                (child,) = node.children
                child.update_parent(node.parent)
                child.branch_length += node.branch_length  # pyright: ignore
                if node.parent is None:
                    return child
                else:
                    node.parent.remove_child(node)
        return tree
